/*
 * Licensed under a Creative Commons Attribution 2.5 Slovenia License
 * http://creativecommons.org/licenses/by/2.5/si/
 * 2009 TineL Studio
 */

package net.tinelstudio.gis.routing.service;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeoutException;

import net.tinelstudio.gis.common.NotEnabledException;
import net.tinelstudio.gis.common.ServiceException;
import net.tinelstudio.gis.common.dto.StreetDto.Level;
import net.tinelstudio.gis.common.service.AbstractEnablableService;
import net.tinelstudio.gis.model.dao.FindingDao;
import net.tinelstudio.gis.model.domain.StreetNode;
import net.tinelstudio.gis.routing.dto.Route;
import net.tinelstudio.gis.routing.searchalgorithm.astar.DefaultAStarSearchAlgorithm;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.GreatCircleDistanceHeuristicFunction;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.Heuristic;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.HeuristicFunction;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.HeuristicFunctionFactory;
import net.tinelstudio.gis.routing.searchalgorithm.cost.Cost;
import net.tinelstudio.gis.routing.searchalgorithm.cost.CostFunction;
import net.tinelstudio.gis.routing.searchalgorithm.cost.CostFunctionFactory;
import net.tinelstudio.gis.routing.searchalgorithm.cost.LevelLengthCostFunction;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import org.springframework.util.StopWatch;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.PrecisionModel;

/**
 * The server side implementation of the Routing service.<br>
 * Uses {@link DefaultAStarSearchAlgorithm}.
 * 
 * @author TineL
 */
public class RoutingServiceImpl extends AbstractEnablableService implements
  RoutingService {

  private final Log logger=LogFactory.getLog(getClass());

  private int timeout=60000;

  private int streetNodeMaxDistanceMeters=10000;

  private FindingDao findingDao;

  private CostFunctionFactory costFunctionFactory;

  private HeuristicFunctionFactory heuristicFunctionFactory;

  private GeometryFactory geometryFactory;

  /**
   * Creates a default routing search algorithm.
   * 
   * @return a default routing search algorithm (<code>null</code> not possible)
   */
  protected DefaultAStarSearchAlgorithm createSearchAlgorithm() {
    return new DefaultAStarSearchAlgorithm();
  }

  protected Route findRoute(Coordinate startLocation, Coordinate goalLocation,
    CostFunction costFunction, HeuristicFunction heuristicFunction)
    throws TimeoutException {

    if (startLocation.equals(goalLocation)) {
      if (logger.isInfoEnabled()) {
        logger.info("Start location is same as goal location. No route needed");
      }
      return null;
    }

    // Find nodes
    StreetNode startNode=findNearestStreetNode(startLocation);
    StreetNode goalNode=findNearestStreetNode(goalLocation);
    if (startNode==null||goalNode==null) {
      return null;
    }

    // Init algorithm
    DefaultAStarSearchAlgorithm aStar=createSearchAlgorithm();
    aStar.setStartNode(startNode);
    aStar.setGoalNode(goalNode);
    aStar.setCostFunction(costFunction);
    aStar.setHeuristicFunction(heuristicFunction);
    aStar.setTimeout(this.timeout);
    aStar.setFindingDao(getFindingDao());

    // Run
    return aStar.run();
  }

  @Transactional(readOnly=true)
  @Override
  public Route findRoute(Coordinate startLocation, Coordinate goalLocation,
    Cost cost, Heuristic heuristic) throws NotEnabledException,
    TimeoutException, ServiceException {
    if (!isEnabled()) throw new NotEnabledException();

    if (logger.isDebugEnabled()) {
      logger
        .debug("Calling Routing service with startLocation="+startLocation
          +", goalLocation="+goalLocation+", cost="+cost+", heuristic="
          +heuristic);
    }

    StopWatch sw=new StopWatch();
    sw.start();

    Route route;
    try {

      CostFunction costFunction=getCostFunctionFactory().createCostFunction(
        cost);
      HeuristicFunction heuristicFunction=getHeuristicFunctionFactory()
        .createHeuristicFunction(heuristic);

      route=findRoute(startLocation, goalLocation, costFunction,
        heuristicFunction);

    } catch (TimeoutException e) {
      throw e;

    } catch (Throwable t) {
      logger.error("", t);
      throw new ServiceException(t);
    }

    sw.stop();
    if (logger.isDebugEnabled()) {
      logger
        .debug("Found route "+route+" in "+sw.getLastTaskTimeMillis()+" ms");
    }

    return route;
  }

  @Transactional(readOnly=true)
  @Override
  public Route findRoute(Coordinate startLocation, Coordinate goalLocation)
    throws NotEnabledException, TimeoutException, ServiceException {
    if (!isEnabled()) throw new NotEnabledException();

    if (logger.isDebugEnabled()) {
      logger.debug("Calling Routing service with startLocation="+startLocation
        +", goalLocation="+goalLocation);
    }

    StopWatch sw=new StopWatch();
    sw.start();

    Route route;
    try {

      // Create default level/weight mapped cost function
      Map<Level, Double> speedMap=new HashMap<Level, Double>();
      speedMap.put(Level.UNKNOWN, 90/70.0);
      speedMap.put(Level.MAJOR, 90/110.0);
      speedMap.put(Level.MEDIUM, 90/65.0);
      speedMap.put(Level.MINOR, 90/55.0);
      CostFunction costFunction=new LevelLengthCostFunction(speedMap);

      // Create default heuristic function
      HeuristicFunction heuristicFunction=new GreatCircleDistanceHeuristicFunction(
        1.24);

      // Find
      route=findRoute(startLocation, goalLocation, costFunction,
        heuristicFunction);

    } catch (TimeoutException e) {
      throw e;

    } catch (Throwable t) {
      logger.error("", t);
      throw new ServiceException(t);
    }

    sw.stop();
    if (logger.isDebugEnabled()) {
      logger
        .debug("Found route "+route+" in "+sw.getLastTaskTimeMillis()+" ms");
    }

    return route;
  }

  /**
   * Finds the nearest street node in the DB.
   * 
   * @param location the location (<code>null</code> not permitted)
   * @return a found street node or <code>null</code> if none found
   */
  protected StreetNode findNearestStreetNode(Coordinate location) {
    long st=System.currentTimeMillis();

    Point point=getGeometryFactory().createPoint(location);
    StreetNode node=getFindingDao().findNearestStreetNode(point,
      getStreetNodeMaxDistanceMeters());

    long et=System.currentTimeMillis();

    if (node==null&&logger.isWarnEnabled()) {
      logger.warn("No node found for "+location+" in "+(et-st)+" ms");

    } else if (logger.isInfoEnabled()) {
      logger.info("Node "+node+" found in "+(et-st)+" ms");
    }

    return node;
  }

  @Override
  public int getTimeout() throws NotEnabledException {
    if (!isEnabled()) throw new NotEnabledException();
    return this.timeout;
  }

  @Override
  public long getServerTime() {
    return System.currentTimeMillis();
  }

  // ########################################

  public CostFunctionFactory getCostFunctionFactory() {
    if (this.costFunctionFactory==null) {
      this.costFunctionFactory=new CostFunctionFactory();
    }
    return this.costFunctionFactory;
  }

  public void setCostFunctionFactory(CostFunctionFactory costFunctionFactory) {
    this.costFunctionFactory=costFunctionFactory;
  }

  public HeuristicFunctionFactory getHeuristicFunctionFactory() {
    if (this.heuristicFunctionFactory==null) {
      this.heuristicFunctionFactory=new HeuristicFunctionFactory();
    }
    return this.heuristicFunctionFactory;
  }

  public void setHeuristicFunctionFactory(
    HeuristicFunctionFactory heuristicFunctionFactory) {
    this.heuristicFunctionFactory=heuristicFunctionFactory;
  }

  public GeometryFactory getGeometryFactory() {
    if (this.geometryFactory==null) {
      this.geometryFactory=new GeometryFactory(new PrecisionModel(), 4326);
    }
    return this.geometryFactory;
  }

  public void setGeometryFactory(GeometryFactory geometryFactory) {
    this.geometryFactory=geometryFactory;
  }

  public void setTimeout(int timeout) {
    this.timeout=timeout;
  }

  public int getStreetNodeMaxDistanceMeters() {
    return this.streetNodeMaxDistanceMeters;
  }

  /**
   * @param streetNodeMaxDistanceMeters the search radius for street nodes in
   *        meters (must be > 0)
   */
  public void setStreetNodeMaxDistanceMeters(int streetNodeMaxDistanceMeters) {
    Assert.isTrue(streetNodeMaxDistanceMeters>0,
      "streetNodeMaxDistanceMeters cannot be <= 0");
    this.streetNodeMaxDistanceMeters=streetNodeMaxDistanceMeters;
  }

  public FindingDao getFindingDao() {
    return this.findingDao;
  }

  public void setFindingDao(FindingDao findingDao) {
    this.findingDao=findingDao;
  }
}
